#ifndef OPERATOR_H
#define OPERATOR_H

#include "globals.h"
#include "state.h"
#include "utilities.h"

#include <cassert>
#include <iostream>
#include <map>
#include <string>
#include <vector>

struct Condition {
    int var;
    int val;
    explicit Condition(std::istream &in);
    Condition(int v, int p) : var(v), val(p) {
        assert(in_bounds(var, g_variable_name));
        assert(val >= 0 && val < g_variable_domain[var]);
    }

    bool is_applicable(const State &state) const;

    bool operator==(const Condition &other) const {
        return var == other.var && val == other.val;
    }

    bool operator!=(const Condition &other) const {
        return !(*this == other);
    }

    void dump() const;
};

struct Effect {
    int var;
    int val;
    std::vector<Condition> conditions;
    Effect(int var, int val, const std::vector<Condition> &co)
        : var(var), val(val), conditions(co) {}

    bool does_fire(const State &state) const {
        for (size_t i = 0; i < conditions.size(); ++i){
            if (!conditions[i].is_applicable(state)){
                return false;
            }
        }
        return true;
    }

    void dump() const;
};

class Operator {
    bool is_an_axiom;
    std::vector<Condition> preconditions;      // var, val
    std::vector<Effect> effects;     // var, val, effect conditions
    
    std::vector<const Condition*> red_pre;
    std::vector<const Condition*> black_pre;
    std::vector<const Effect*> red_eff;
    std::vector<const Effect*> black_eff;
    
    std::string name;
    int cost;
    
    // has effect on black variable?
    bool is_black_operator;

    mutable bool marked; // Used for short-term marking of preferred operators
    
    void read_pre_post(std::istream &in);
public:
    explicit Operator(std::istream &in, bool is_axiom);
    void dump() const;
    std::string get_name() const {return name; }

    bool is_axiom() const {return is_an_axiom; }
    
    const std::vector<Condition> &get_preconditions() const {
        return preconditions;
    }
    
    const std::vector<Effect> &get_effects() const {
        return effects;
    }
    
    const std::vector<const Condition*> get_red_pre() const {
        return red_pre;
    }
    
    const std::vector<const Effect*> get_red_eff() const {
        return red_eff;
    }
    
    const std::vector<const Condition*> get_black_pre() const {
        return black_pre;
    }
    
    const std::vector<const Effect*> get_black_eff() const {
        return black_eff;
    }
    
    bool is_applicable(const State &state) const;
    
    bool is_marked() const {
        return marked;
    }
    
    void mark() const {
        marked = true;
    }
    
    bool is_black_op() const {
        return is_black_operator;
    }
    
    void do_paint();
    
    void unmark() const {
        marked = false;
    }

    int get_cost() const {
        return cost;
    }
};

#endif
